#include <assert.h>

#include <aterm2.h>
#include <deprecated.h>
#include "rolodex.h"

/*{{{  conversion functions */

ATerm stringToChars(const char *str)
{
  int len = strlen(str);
  int i;
  ATermList result = ATempty;

  for (i = len - 1; i >= 0; i--) {
    result = ATinsert(result, (ATerm) ATmakeInt(str[i]));
  }

  return (ATerm) result;
}

char *charsToString(ATerm arg)
{
  ATermList list = (ATermList) arg;
  int len = ATgetLength(list);
  int i;
  char *str;

  str = (char *) malloc(len+1);
  if (str == NULL) {
      return NULL;
  }

  for (i = 0; !ATisEmpty(list); list = ATgetNext(list), i++) {
    str[i] = (char) ATgetInt((ATermInt) ATgetFirst(list));
  }
  str[i] = '\0';

  return str;
}


/*}}}  */

/*{{{  typedefs */

typedef struct ATerm _RoloList;
typedef struct ATerm _Rolodex;
typedef struct ATerm _PhoneNumber;

/*}}}  */

/*{{{  void initRolodexApi(void) */

void initRolodexApi(void)
{
  init_rolodex_dict();
}

/*}}}  */

/*{{{  protect functions */

void protectRoloList(RoloList *arg)
{
  ATprotect((ATerm*)((void*) arg));
}

void protectRolodex(Rolodex *arg)
{
  ATprotect((ATerm*)((void*) arg));
}

void protectPhoneNumber(PhoneNumber *arg)
{
  ATprotect((ATerm*)((void*) arg));
}


/*}}}  */
/*{{{  term conversion functions */

/*{{{  RoloList RoloListFromTerm(ATerm t) */

RoloList RoloListFromTerm(ATerm t)
{
  return (RoloList)t;
}

/*}}}  */
/*{{{  ATerm RoloListToTerm(RoloList arg) */

ATerm RoloListToTerm(RoloList arg)
{
  return (ATerm)arg;
}

/*}}}  */
/*{{{  Rolodex RolodexFromTerm(ATerm t) */

Rolodex RolodexFromTerm(ATerm t)
{
  return (Rolodex)t;
}

/*}}}  */
/*{{{  ATerm RolodexToTerm(Rolodex arg) */

ATerm RolodexToTerm(Rolodex arg)
{
  return (ATerm)arg;
}

/*}}}  */
/*{{{  PhoneNumber PhoneNumberFromTerm(ATerm t) */

PhoneNumber PhoneNumberFromTerm(ATerm t)
{
  return (PhoneNumber)t;
}

/*}}}  */
/*{{{  ATerm PhoneNumberToTerm(PhoneNumber arg) */

ATerm PhoneNumberToTerm(PhoneNumber arg)
{
  return (ATerm)arg;
}

/*}}}  */

/*}}}  */
/*{{{  list functions */

int getRoloListLength (RoloList arg) {
  return ATgetLength((ATermList) arg);
}
RoloList reverseRoloList(RoloList arg) {
  return (RoloList) ATreverse((ATermList) arg);
}
RoloList appendRoloList(RoloList arg, Rolodex elem) {
  return (RoloList) ATappend((ATermList) arg, (ATerm) ((ATerm) elem));
}
RoloList concatRoloList(RoloList arg0, RoloList arg1) {
  return (RoloList) ATconcat((ATermList) arg0, (ATermList) arg1);
}
RoloList sliceRoloList(RoloList arg, int start, int end) {
  return (RoloList) ATgetSlice((ATermList) arg, start, end);
}
Rolodex getRoloListRolodexAt(RoloList arg, int index) {
 return (Rolodex)ATelementAt((ATermList) arg,index);
}
RoloList replaceRoloListRolodexAt(RoloList arg, Rolodex elem, int index) {
 return (RoloList) ATreplace((ATermList) arg, (ATerm) ((ATerm) elem), index);
}
RoloList makeRoloList2(Rolodex elem1, Rolodex elem2) {
  return (RoloList) ATmakeList2((ATerm) ((ATerm) elem2), (ATerm) ((ATerm) elem2));
}
RoloList makeRoloList3(Rolodex elem1, Rolodex elem2, Rolodex elem3) {
  return (RoloList) ATmakeList3((ATerm) ((ATerm) elem3), (ATerm) ((ATerm) elem3), (ATerm) ((ATerm) elem3));
}
RoloList makeRoloList4(Rolodex elem1, Rolodex elem2, Rolodex elem3, Rolodex elem4) {
  return (RoloList) ATmakeList4((ATerm) ((ATerm) elem4), (ATerm) ((ATerm) elem4), (ATerm) ((ATerm) elem4), (ATerm) ((ATerm) elem4));
}
RoloList makeRoloList5(Rolodex elem1, Rolodex elem2, Rolodex elem3, Rolodex elem4, Rolodex elem5) {
  return (RoloList) ATmakeList5((ATerm) ((ATerm) elem5), (ATerm) ((ATerm) elem5), (ATerm) ((ATerm) elem5), (ATerm) ((ATerm) elem5), (ATerm) ((ATerm) elem5));
}
RoloList makeRoloList6(Rolodex elem1, Rolodex elem2, Rolodex elem3, Rolodex elem4, Rolodex elem5, Rolodex elem6) {
  return (RoloList) ATmakeList6((ATerm) ((ATerm) elem6), (ATerm) ((ATerm) elem6), (ATerm) ((ATerm) elem6), (ATerm) ((ATerm) elem6), (ATerm) ((ATerm) elem6), (ATerm) ((ATerm) elem6));
}

/*}}}  */
/*{{{  constructors */

/*{{{  RoloList makeRoloListEmpty(void) */

RoloList makeRoloListEmpty(void)
{
  return (RoloList)(ATerm)ATempty;
}

/*}}}  */
/*{{{  RoloList makeRoloListSingle(Rolodex head) */

RoloList makeRoloListSingle(Rolodex head)
{
  return (RoloList)(ATerm)ATmakeList1((ATerm) head);
}

/*}}}  */
/*{{{  RoloList makeRoloListMany(Rolodex head, RoloList tail) */

RoloList makeRoloListMany(Rolodex head, RoloList tail)
{
  return (RoloList)(ATerm)ATinsert((ATermList)tail, (ATerm) head);
}

/*}}}  */
/*{{{  Rolodex makeRolodexHome(const char* name, PhoneNumber phone) */

Rolodex makeRolodexHome(const char* name, PhoneNumber phone)
{
  return (Rolodex)(ATerm)ATmakeAppl1(afun0, (ATerm)ATinsert(ATmakeList1((ATerm) phone), (ATerm) (ATerm) ATmakeAppl(ATmakeAFun(name, 0, ATtrue))));
}

/*}}}  */
/*{{{  Rolodex makeRolodexWork(const char* company, PhoneNumber phone) */

Rolodex makeRolodexWork(const char* company, PhoneNumber phone)
{
  return (Rolodex)(ATerm)ATmakeAppl1(afun1, (ATerm)ATinsert(ATmakeList1((ATerm) phone), (ATerm) (ATerm) ATmakeAppl(ATmakeAFun(company, 0, ATtrue))));
}

/*}}}  */
/*{{{  PhoneNumber makePhoneNumberVoice(int voice) */

PhoneNumber makePhoneNumberVoice(int voice)
{
  return (PhoneNumber)(ATerm)ATmakeAppl1(afun2, (ATerm) (ATerm) ATmakeInt(voice));
}

/*}}}  */
/*{{{  PhoneNumber makePhoneNumberFax(int fax) */

PhoneNumber makePhoneNumberFax(int fax)
{
  return (PhoneNumber)(ATerm)ATmakeAppl1(afun3, (ATerm) (ATerm) ATmakeInt(fax));
}

/*}}}  */

/*}}}  */
/*{{{  equality functions */

ATbool isEqualRoloList(RoloList arg0, RoloList arg1)
{
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

ATbool isEqualRolodex(Rolodex arg0, Rolodex arg1)
{
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

ATbool isEqualPhoneNumber(PhoneNumber arg0, PhoneNumber arg1)
{
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/*}}}  */
/*{{{  RoloList accessors */

/*{{{  ATbool isValidRoloList(RoloList arg) */

ATbool isValidRoloList(RoloList arg)
{
  if (isRoloListEmpty(arg)) {
    return ATtrue;
  }
  else if (isRoloListSingle(arg)) {
    return ATtrue;
  }
  else if (isRoloListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/*}}}  */
/*{{{  inline ATbool isRoloListEmpty(RoloList arg) */

inline ATbool isRoloListEmpty(RoloList arg)
{
  if (!ATisEmpty((ATermList)arg)) {
    return ATfalse;
  }
#ifndef DISABLE_DYNAMIC_CHECKING
  assert(arg != NULL);
  assert(ATmatchTerm((ATerm)arg, patternRoloListEmpty));
#endif
  return ATtrue;
}

/*}}}  */
/*{{{  inline ATbool isRoloListSingle(RoloList arg) */

inline ATbool isRoloListSingle(RoloList arg)
{
  if (ATisEmpty((ATermList)arg)) {
    return ATfalse;
  }
  {
    static ATerm last_arg = NULL;
    static int last_gc = -1;
    static ATbool last_result;

    assert(arg != NULL);

    if (last_gc != ATgetGCCount() || (ATerm)arg != last_arg) {
      last_arg = (ATerm)arg;
      last_result = ATmatchTerm((ATerm)arg, patternRoloListSingle, NULL);
      last_gc = ATgetGCCount();
    }

    return last_result;
  }
}

/*}}}  */
/*{{{  inline ATbool isRoloListMany(RoloList arg) */

inline ATbool isRoloListMany(RoloList arg)
{
  if (ATisEmpty((ATermList)arg)) {
    return ATfalse;
  }
  {
    static ATerm last_arg = NULL;
    static int last_gc = -1;
    static ATbool last_result;

    assert(arg != NULL);

    if (last_gc != ATgetGCCount() || (ATerm)arg != last_arg) {
      last_arg = (ATerm)arg;
      last_result = ATmatchTerm((ATerm)arg, patternRoloListMany, NULL, NULL);
      last_gc = ATgetGCCount();
    }

    return last_result;
  }
}

/*}}}  */
/*{{{  ATbool hasRoloListHead(RoloList arg) */

ATbool hasRoloListHead(RoloList arg)
{
  if (isRoloListSingle(arg)) {
    return ATtrue;
  }
  else if (isRoloListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/*}}}  */
/*{{{  ATbool hasRoloListTail(RoloList arg) */

ATbool hasRoloListTail(RoloList arg)
{
  if (isRoloListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/*}}}  */
/*{{{  Rolodex getRoloListHead(RoloList arg) */

Rolodex getRoloListHead(RoloList arg)
{
  if (isRoloListSingle(arg)) {
    return (Rolodex)ATgetFirst((ATermList)arg);
  }
  else 
    return (Rolodex)ATgetFirst((ATermList)arg);
}

/*}}}  */
/*{{{  RoloList getRoloListTail(RoloList arg) */

RoloList getRoloListTail(RoloList arg)
{
  
    return (RoloList)ATgetNext((ATermList)arg);
}

/*}}}  */
/*{{{  RoloList setRoloListHead(RoloList arg, Rolodex head) */

RoloList setRoloListHead(RoloList arg, Rolodex head)
{
  if (isRoloListSingle(arg)) {
    return (RoloList)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }
  else if (isRoloListMany(arg)) {
    return (RoloList)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }

  ATabort("RoloList has no Head: %t\n", arg);
  return (RoloList)NULL;
}

/*}}}  */
/*{{{  RoloList setRoloListTail(RoloList arg, RoloList tail) */

RoloList setRoloListTail(RoloList arg, RoloList tail)
{
  if (isRoloListMany(arg)) {
    return (RoloList)ATreplaceTail((ATermList)arg, (ATermList)((ATerm) tail), 1);
  }

  ATabort("RoloList has no Tail: %t\n", arg);
  return (RoloList)NULL;
}

/*}}}  */

/*}}}  */
/*{{{  Rolodex accessors */

/*{{{  ATbool isValidRolodex(Rolodex arg) */

ATbool isValidRolodex(Rolodex arg)
{
  if (isRolodexHome(arg)) {
    return ATtrue;
  }
  else if (isRolodexWork(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/*}}}  */
/*{{{  inline ATbool isRolodexHome(Rolodex arg) */

inline ATbool isRolodexHome(Rolodex arg)
{
  {
    static ATerm last_arg = NULL;
    static int last_gc = -1;
    static ATbool last_result;

    assert(arg != NULL);

    if (last_gc != ATgetGCCount() || (ATerm)arg != last_arg) {
      last_arg = (ATerm)arg;
      last_result = ATmatchTerm((ATerm)arg, patternRolodexHome, NULL, NULL);
      last_gc = ATgetGCCount();
    }

    return last_result;
  }
}

/*}}}  */
/*{{{  inline ATbool isRolodexWork(Rolodex arg) */

inline ATbool isRolodexWork(Rolodex arg)
{
  {
    static ATerm last_arg = NULL;
    static int last_gc = -1;
    static ATbool last_result;

    assert(arg != NULL);

    if (last_gc != ATgetGCCount() || (ATerm)arg != last_arg) {
      last_arg = (ATerm)arg;
      last_result = ATmatchTerm((ATerm)arg, patternRolodexWork, NULL, NULL);
      last_gc = ATgetGCCount();
    }

    return last_result;
  }
}

/*}}}  */
/*{{{  ATbool hasRolodexName(Rolodex arg) */

ATbool hasRolodexName(Rolodex arg)
{
  if (isRolodexHome(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/*}}}  */
/*{{{  ATbool hasRolodexPhone(Rolodex arg) */

ATbool hasRolodexPhone(Rolodex arg)
{
  if (isRolodexHome(arg)) {
    return ATtrue;
  }
  else if (isRolodexWork(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/*}}}  */
/*{{{  ATbool hasRolodexCompany(Rolodex arg) */

ATbool hasRolodexCompany(Rolodex arg)
{
  if (isRolodexWork(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/*}}}  */
/*{{{  char* getRolodexName(Rolodex arg) */

char* getRolodexName(Rolodex arg)
{
  
    return (char*)ATgetName(ATgetAFun((ATermAppl) ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 0))));
}

/*}}}  */
/*{{{  PhoneNumber getRolodexPhone(Rolodex arg) */

PhoneNumber getRolodexPhone(Rolodex arg)
{
  if (isRolodexHome(arg)) {
    return (PhoneNumber)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 0), 1);
  }
  else 
    return (PhoneNumber)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 0), 1);
}

/*}}}  */
/*{{{  char* getRolodexCompany(Rolodex arg) */

char* getRolodexCompany(Rolodex arg)
{
  
    return (char*)ATgetName(ATgetAFun((ATermAppl) ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 0))));
}

/*}}}  */
/*{{{  Rolodex setRolodexName(Rolodex arg, const char* name) */

Rolodex setRolodexName(Rolodex arg, const char* name)
{
  if (isRolodexHome(arg)) {
    return (Rolodex)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 0), (ATerm)((ATerm) (ATerm) ATmakeAppl(ATmakeAFun(name, 0, ATtrue))), 0), 0);
  }

  ATabort("Rolodex has no Name: %t\n", arg);
  return (Rolodex)NULL;
}

/*}}}  */
/*{{{  Rolodex setRolodexPhone(Rolodex arg, PhoneNumber phone) */

Rolodex setRolodexPhone(Rolodex arg, PhoneNumber phone)
{
  if (isRolodexHome(arg)) {
    return (Rolodex)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 0), (ATerm)((ATerm) phone), 1), 0);
  }
  else if (isRolodexWork(arg)) {
    return (Rolodex)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 0), (ATerm)((ATerm) phone), 1), 0);
  }

  ATabort("Rolodex has no Phone: %t\n", arg);
  return (Rolodex)NULL;
}

/*}}}  */
/*{{{  Rolodex setRolodexCompany(Rolodex arg, const char* company) */

Rolodex setRolodexCompany(Rolodex arg, const char* company)
{
  if (isRolodexWork(arg)) {
    return (Rolodex)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 0), (ATerm)((ATerm) (ATerm) ATmakeAppl(ATmakeAFun(company, 0, ATtrue))), 0), 0);
  }

  ATabort("Rolodex has no Company: %t\n", arg);
  return (Rolodex)NULL;
}

/*}}}  */

/*}}}  */
/*{{{  PhoneNumber accessors */

/*{{{  ATbool isValidPhoneNumber(PhoneNumber arg) */

ATbool isValidPhoneNumber(PhoneNumber arg)
{
  if (isPhoneNumberVoice(arg)) {
    return ATtrue;
  }
  else if (isPhoneNumberFax(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/*}}}  */
/*{{{  inline ATbool isPhoneNumberVoice(PhoneNumber arg) */

inline ATbool isPhoneNumberVoice(PhoneNumber arg)
{
  {
    static ATerm last_arg = NULL;
    static int last_gc = -1;
    static ATbool last_result;

    assert(arg != NULL);

    if (last_gc != ATgetGCCount() || (ATerm)arg != last_arg) {
      last_arg = (ATerm)arg;
      last_result = ATmatchTerm((ATerm)arg, patternPhoneNumberVoice, NULL);
      last_gc = ATgetGCCount();
    }

    return last_result;
  }
}

/*}}}  */
/*{{{  inline ATbool isPhoneNumberFax(PhoneNumber arg) */

inline ATbool isPhoneNumberFax(PhoneNumber arg)
{
  {
    static ATerm last_arg = NULL;
    static int last_gc = -1;
    static ATbool last_result;

    assert(arg != NULL);

    if (last_gc != ATgetGCCount() || (ATerm)arg != last_arg) {
      last_arg = (ATerm)arg;
      last_result = ATmatchTerm((ATerm)arg, patternPhoneNumberFax, NULL);
      last_gc = ATgetGCCount();
    }

    return last_result;
  }
}

/*}}}  */
/*{{{  ATbool hasPhoneNumberVoice(PhoneNumber arg) */

ATbool hasPhoneNumberVoice(PhoneNumber arg)
{
  if (isPhoneNumberVoice(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/*}}}  */
/*{{{  ATbool hasPhoneNumberFax(PhoneNumber arg) */

ATbool hasPhoneNumberFax(PhoneNumber arg)
{
  if (isPhoneNumberFax(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/*}}}  */
/*{{{  int getPhoneNumberVoice(PhoneNumber arg) */

int getPhoneNumberVoice(PhoneNumber arg)
{
  
    return (int)ATgetInt((ATermInt) ATgetArgument((ATermAppl)arg, 0));
}

/*}}}  */
/*{{{  int getPhoneNumberFax(PhoneNumber arg) */

int getPhoneNumberFax(PhoneNumber arg)
{
  
    return (int)ATgetInt((ATermInt) ATgetArgument((ATermAppl)arg, 0));
}

/*}}}  */
/*{{{  PhoneNumber setPhoneNumberVoice(PhoneNumber arg, int voice) */

PhoneNumber setPhoneNumberVoice(PhoneNumber arg, int voice)
{
  if (isPhoneNumberVoice(arg)) {
    return (PhoneNumber)ATsetArgument((ATermAppl)arg, (ATerm)((ATerm) (ATerm) ATmakeInt(voice)), 0);
  }

  ATabort("PhoneNumber has no Voice: %t\n", arg);
  return (PhoneNumber)NULL;
}

/*}}}  */
/*{{{  PhoneNumber setPhoneNumberFax(PhoneNumber arg, int fax) */

PhoneNumber setPhoneNumberFax(PhoneNumber arg, int fax)
{
  if (isPhoneNumberFax(arg)) {
    return (PhoneNumber)ATsetArgument((ATermAppl)arg, (ATerm)((ATerm) (ATerm) ATmakeInt(fax)), 0);
  }

  ATabort("PhoneNumber has no Fax: %t\n", arg);
  return (PhoneNumber)NULL;
}

/*}}}  */

/*}}}  */
/*{{{  sort visitors */

/*{{{  RoloList visitRoloList(RoloList arg, Rolodex (*acceptHead)(Rolodex)) */

RoloList visitRoloList(RoloList arg, Rolodex (*acceptHead)(Rolodex))
{
  if (isRoloListEmpty(arg)) {
    return makeRoloListEmpty();
  }
  if (isRoloListSingle(arg)) {
    return makeRoloListSingle(
        acceptHead ? acceptHead(getRoloListHead(arg)) : getRoloListHead(arg));
  }
  if (isRoloListMany(arg)) {
    return makeRoloListMany(
        acceptHead ? acceptHead(getRoloListHead(arg)) : getRoloListHead(arg),
        visitRoloList(getRoloListTail(arg), acceptHead));
  }
  ATabort("not a RoloList: %t\n", arg);
  return (RoloList)NULL;
}

/*}}}  */
/*{{{  Rolodex visitRolodex(Rolodex arg, char* (*acceptName)(char*), PhoneNumber (*acceptPhone)(PhoneNumber), char* (*acceptCompany)(char*)) */

Rolodex visitRolodex(Rolodex arg, char* (*acceptName)(char*), PhoneNumber (*acceptPhone)(PhoneNumber), char* (*acceptCompany)(char*))
{
  if (isRolodexHome(arg)) {
    return makeRolodexHome(
        acceptName ? acceptName(getRolodexName(arg)) : getRolodexName(arg),
        acceptPhone ? acceptPhone(getRolodexPhone(arg)) : getRolodexPhone(arg));
  }
  if (isRolodexWork(arg)) {
    return makeRolodexWork(
        acceptCompany ? acceptCompany(getRolodexCompany(arg)) : getRolodexCompany(arg),
        acceptPhone ? acceptPhone(getRolodexPhone(arg)) : getRolodexPhone(arg));
  }
  ATabort("not a Rolodex: %t\n", arg);
  return (Rolodex)NULL;
}

/*}}}  */
/*{{{  PhoneNumber visitPhoneNumber(PhoneNumber arg, int (*acceptVoice)(int), int (*acceptFax)(int)) */

PhoneNumber visitPhoneNumber(PhoneNumber arg, int (*acceptVoice)(int), int (*acceptFax)(int))
{
  if (isPhoneNumberVoice(arg)) {
    return makePhoneNumberVoice(
        acceptVoice ? acceptVoice(getPhoneNumberVoice(arg)) : getPhoneNumberVoice(arg));
  }
  if (isPhoneNumberFax(arg)) {
    return makePhoneNumberFax(
        acceptFax ? acceptFax(getPhoneNumberFax(arg)) : getPhoneNumberFax(arg));
  }
  ATabort("not a PhoneNumber: %t\n", arg);
  return (PhoneNumber)NULL;
}

/*}}}  */

/*}}}  */
